#!/usr/bin/env python3

import sys

"""
obja parser for python.
"""

class Vertex:
    """
    The class that holds the x, y, and z coordinates of a vertex.
    """
    def __init__(self, array):
        """
        Initializes a vertex from an array of string representing floats.
        """
        self.set(array)

    def set(self, array):
        """
        Sets a vertex from an array of string representing floats.
        """
        self.x = float(array[0])
        self.y = float(array[1])
        self.z = float(array[2])

    def translate(self, array):
        """
        Translates a vertex from an array of string representing floats.
        """
        self.x += float(array[0])
        self.y += float(array[1])
        self.z += float(array[2])

class Face:
    """
    The class that holds a, b, and c, the indices of the vertices of the face.
    """
    def __init__(self, array):
        """
        Initializes a face from an array of strings representing vertex indices (starting at 1)
        """
        self.set(array)
        self.visible = True

    def set(self, array):
        """
        Sets a face from an array of strings representing vertex indices (starting at 1)
        """
        self.a = int(array[0]) - 1
        self.b = int(array[1]) - 1
        self.c = int(array[2]) - 1

    def test(self, vertices, line = "unknown"):
        """
        Tests if a face references only vertices that exist when the face is declared.
        """
        if self.a >= len(vertices):
            raise VertexError(self.a + 1, line)
        if self.b >= len(vertices):
            raise VertexError(self.b + 1, line)
        if self.c >= len(vertices):
            raise VertexError(self.c + 1, line)

class VertexError(Exception):
    """
    An operation references a vertex that does not exist.
    """
    def __init__(self, index, line):
        """
        Creates the error from index of the referenced vertex and the line where the error occured.
        """
        self.line = line
        self.index = index
        super().__init__()

    def __str__(self):
        """
        Pretty prints the error.
        """
        return f'There is no vertex {self.index} (line {self.line})'

class FaceError(Exception):
    """
    An operation references a face that does not exist.
    """
    def __init__(self, index, line):
        """
        Creates the error from index of the referenced face and the line where the error occured.
        """
        self.line = line
        self.index = index
        super().__init__()

    def __str__(self):
        """
        Pretty prints the error.
        """
        return f'There is no face {self.index} (line {self.line})'

class FaceVertexError(Exception):
    """
    An operation references a face vertex that does not exist.
    """
    def __init__(self, index, line):
        """
        Creates the error from index of the referenced face vertex and the line where the error occured.
        """
        self.line = line
        self.index = index
        super().__init__()

    def __str__(self):
        """
        Pretty prints the error.
        """
        return f'Face has no vertex {self.index} (line {self.line})'

class UnknownInstruction(Exception):
    """
    An instruction is unknown.
    """
    def __init__(self, instruction, line):
        """
        Creates the error from instruction and the line where the error occured.
        """
        self.line = line
        self.instruction = instruction
        super().__init__()

    def __str__(self):
        """
        Pretty prints the error.
        """
        return f'Instruction {self.instruction} unknown (line {self.line})'

class Parser:
    """
    The OBJA parser.
    """
    def __init__(self):
        """
        Intializes an empty parser.
        """
        self.vertices = []
        self.faces = []
        self.line = 0

    def get_vertex(self, string):
        """
        Gets a vertex from a string representing the index of the vertex, starting at 1.
        """
        index = int(string) - 1
        if index >= len(self.vertices):
            raise FaceError(index + 1, self.line)
        return self.vertices[index]

    def get_face(self, string):
        """
        Gets a face from a string representing the index of the face, starting at 1.
        """
        index = int(string) - 1
        if index >= len(self.faces):
            raise FaceError(index + 1, self.line)
        return self.faces[index]

    def parse_line(self, line):
        """
        Parses a line of obja file.
        """
        self.line += 1

        split = line.split()

        if len(split) == 0:
            return

        if split[0] == "v":
            self.vertices.append(Vertex(split[1:]))

        elif split[0] == "ev":
            self.get_vertex(split[1]).set(split[2:])

        elif split[0] == "tv":
            self.get_vertex(split[1]).translate(split[2:])

        elif split[0] == "f":
            face = Face(split[1:])
            face.test(self.vertices, self.line)
            self.faces.append(face)

        elif split[0] == "ts":
            for i in range(1, len(split) - 2):
                if i % 2 == 1:
                    face = Face([split[i], split[i + 1], split[i + 2]])
                else:
                    face = Face([split[i], split[i + 2], split[i + 1]])
                face.test(self.vertices, self.line)
                self.faces.append(face)

        elif split[0] == "tf":
            for i in range(1, len(split) - 2):
                face = Face(split[i:i+3])
                face.test(self.vertices, self.line)
                self.faces.append(face)

        elif split[0] == "ef":
            self.get_face(split[1]).set(split[2:])

        elif split[0] == "efv":
            face = self.get_face(split[1])
            vertex = int(split[2])
            new_index = int(split[3]) - 1
            if vertex == 1:
                face.a = new_index
            elif vertex == 2:
                face.b = new_index
            elif vertex == 3:
                face.c = new_index
            else:
                raise FaceVertexError(vertex, self.line)

        elif split[0] == "df":
            self.get_face(split[1]).visible = False

        elif split[0] == "#":
            return

        else:
            raise UnknownInstruction(split[0], self.line)

def main():
    if len(sys.argv) == 1:
        print("obja needs a path to an obja file")
        return

    parser = Parser()
    with open(sys.argv[1], "r") as file:
        for line in file.readlines():
            parser.parse_line(line)

if __name__ == "__main__":
    main()
